#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  SPDX-License-Identifier: GPL-3.0-only
#  Copyright 2022 drad <sa@adercon.com>

import json
import subprocess  # nosec
import time
from operator import itemgetter
from pathlib import Path
from typing import List

import click
import pycouchdb
import requests
import typer
from natsort import natsorted
from rich import box
from rich.table import Table

from mpbroker.config.config import user_cfg
from mpbroker.models.media import (
    MediaMetadata,
    MediaPlayHistory,
    MediaPlayRating,
    MediaPlayStatus,
)

server = pycouchdb.Server(user_cfg.database.db_uri)


def db_not_available():
    """
    Database is not available: display message and exit.
    """

    typer.secho(
        "Database unavailable, is it up?", fg=typer.colors.RED, bold=True, err=True
    )
    raise typer.Exit()


def determine_status_from_flags(
    new: bool = False,
    played: bool = False,
    watched: bool = False,
):
    """
    Determine the status from status flags.
    @return: status to show or None (indicating no filter on status)
    @raises: error if more than one flag set.
    """

    _status = []
    if new:
        _status.append(MediaPlayStatus.new.value)
    if played:
        _status.append(MediaPlayStatus.played.value)
    if watched:
        _status.append(MediaPlayStatus.watched.value)
    if len(_status) < 1:
        return None
    elif len(_status) > 1:
        typer.secho(
            f"Too many status filter flags: {_status}!",
            fg=typer.colors.RED,
            bold=True,
            err=True,
        )
        typer.echo(
            "You can only supply one status to filter on.",
        )
        raise typer.Exit()
    else:
        return _status[0]


def dprint(data):
    """
    Pretty print a dict/list of dicts.
    """

    import pprint

    pprint.pprint(data, compact=True)


def get_sources_paths(sources: List):
    """
    Get sources paths.
    NOTE: source paths are validated (checked), if a path is invalid (does not exist)
          it will not be returned.
    @return: list of source paths.
    """

    _ret = []
    for source in sources:
        # ~ typer.echo(f"   ─⏵ checking source: {source}")
        if source in user_cfg.source_mappings:
            _path = user_cfg.source_mappings[source]
            # check if path exists.
            if Path(_path).is_dir() and any(Path(_path).iterdir()):
                # ~ typer.echo(f"- found {_path} and it has data...")
                _ret.append({"source": source, "path": _path})
            # ~ typer.echo(f"- found {source} in source_mappings: {user_cfg.source_mappings[source]}")

    return _ret


def extract_metadata(filepath) -> MediaMetadata:
    """
    Extract media metadata from a file
    """

    from pymediainfo import MediaInfo

    try:
        media_info = MediaInfo.parse(filepath)

        # extract metadata
        metadata = MediaMetadata(
            file_size=media_info.tracks[0].other_file_size[0],
            file_type=media_info.tracks[1].internet_media_type,
            file_format=media_info.tracks[0].format,
            encoding=media_info.tracks[1].encoded_library_name,
            duration=media_info.tracks[0].other_duration[0],
            resolution=f"{media_info.tracks[1].width} x {media_info.tracks[1].height}",
            aspect_ratio=media_info.tracks[1].other_display_aspect_ratio[0],
            audio_format=media_info.tracks[2].format,
            audio_sampling=media_info.tracks[2].sampling_rate,
        )
    except Exception as e:
        return None, f"Failed extracting metadata for {filepath} with error: {e}"

    return metadata, None


def generate_sid(instr: str = None):
    """
    Generate a sid given an input string.
    Example: 'Duck_Dynasty_S1_D1.mkv' ─⏵ 'DCKDNSTS1D1.MKV'
    """

    return instr.translate(str.maketrans("", "", "AaEeIiOoUuYy_-()[]")).upper()


def jprint(data: str = None):
    """
    JSON pretty print a string.
    """

    # typer.echo(f"- data ({type(data)}): {data}")
    _data = json.loads(data)
    typer.echo(json.dumps(_data, indent=4, sort_keys=True))


def make_doc(doc=None, rename_doc_id: bool = False):
    """
    Create a couchdb doc by deserializing the class to json, loading it back to
    json and finally renaming the doc_id field to _id.

    NOTE: the double deserialization is needed to get dates deserialized (easily)
    NOTE: we need to rename doc_id to _id as we can name it _id in the model or alias
          it as python treats it as a local and does not export it on deserialization.
    """

    j = json.loads(doc.json())
    if "doc_id" in j and rename_doc_id:
        j["_id"] = j.pop("doc_id")

    return j


def play_item(doc=None, name: str = None, user: str = None):
    """
    Play an item.
    """

    db = server.database("media")
    try:
        # ~ typer.echo(f"- playing doc: {doc}")
        # typer.echo(f">> source_mappings: {user_cfg.source_mappings}")
        source_paths = get_sources_paths(doc["sources"])
        _name = f"{doc['directory']}/{doc['name']}"
        if len(source_paths) < 1:
            typer.echo(
                f"No viable sources found for {_name} with sources: {doc['sources']}"
            )
            raise typer.Exit()
        # typer.echo(f" source_paths: {source_paths}")

        # @TODO: currently we simply play the first source_path - should have something
        #  better here eventually like a ranking or some checking.
        _media_path = f"{source_paths[0]['path']}/{_name}"

        # capture when we start playing.
        _start = time.time()

        # @TODO: check if file and player exist, then execute
        subprocess.call([user_cfg.player, _media_path])  # nosec
        _end = time.time()

        # create the play history.
        _new_history = MediaPlayHistory(
            base=source_paths[0]["path"],
            player=user_cfg.player,
            start=_start,
            end=_end,
            client=f"{user}",
        )
        _history = []
        if (
            "play" in doc
            and "history" in doc["play"]
            and doc["play"]["history"]
            and len(doc["play"]["history"]) > 0
        ):
            _history = doc["play"]["history"]

        _history.append(json.loads(_new_history.json()))
        doc["play"]["history"] = _history

        # post-play updates from user
        doc["play"]["status"] = typer.prompt(
            "update Play Status",
            default=MediaPlayStatus.played.value,
            type=click.Choice([str(i) for i in MediaPlayStatus._value2member_map_]),
        )
        doc["play"]["rating"] = int(
            typer.prompt(
                "Rate item",
                default=MediaPlayRating.ok,
                type=click.Choice([str(i) for i in MediaPlayRating._value2member_map_]),
            )
        )
        rating_notes = typer.prompt(
            "Add Rating notes? (leave blank to not add a note)",
            default="",
            show_default=False,
        )
        if rating_notes:
            doc["play"]["rating_notes"] = rating_notes

        notes = typer.prompt(
            "Add Notes for media item? (leave blank to not add a note update)\n  A note for the media item could be something specific like 'watched Ep 1, 2, and 4'",
            default="",
            show_default=False,
        )
        if notes:
            doc["play"]["notes"] = notes

        db.save(doc)

    except requests.exceptions.ConnectionError:
        db_not_available()
    except pycouchdb.exceptions.NotFound:
        typer.secho(
            f"Media item {_name} not found for user {user}.",
            fg=typer.colors.RED,
            bold=True,
        )
        raise typer.Exit()


def results_by_name(
    name: str, user: str, status: str = None, filter_directory: str = ""
):
    """
    Get Results from a given Name.
    """

    _name = f"{user}:{name}"
    _status = f"-{status}" if status else ""

    try:
        db = server.database("media")
        results = db.query(
            f"filters/by-name{_status}",
            startkey=_name,
            endkey=f"{_name}\ufff0",
            as_list=True,
        )
        if results:
            # post process results.
            if filter_directory:
                _results = []
                for item in results:
                    if item["value"][0] and filter_directory in item["value"][0]:
                        _results.append(item)

                if len(_results) > 0:
                    return _results
                else:
                    typer.secho("No results found", fg=typer.colors.MAGENTA, bold=True)
                    raise typer.Exit()

            return results
        else:
            typer.secho("No results found", fg=typer.colors.MAGENTA, bold=True)
            raise typer.Exit()
    except requests.exceptions.ConnectionError:
        db_not_available()


def results_to_table(
    results=None, name: str = None, user: str = None, directory: str = None
):
    """
    Make query results into a [rich] table for display.
    """

    _directory_label = f" in '{directory}'" if directory else ""
    _caption = typer.style(
        f"{len(results)} items found for term '{name}'{_directory_label}",
        fg=typer.colors.MAGENTA,
        bold=False,
    )
    table = Table(
        box=box.ROUNDED,
        show_lines=False,
        caption=_caption,
        collapse_padding=True,
        pad_edge=False,
        padding=0,
        show_edge=True,
        leading=0,
        header_style="bold magenta",
    )
    table.add_column("Item", style="cyan", no_wrap=True)
    table.add_column("Status", justify="center", style="magenta")
    table.add_column("Rating")
    table.add_column("Notes")
    table.add_column("Sources", justify="right", style="yellow")
    table.add_column("Length")

    for item in natsorted(results, key=itemgetter(*["id"])):
        _status = item["value"][1]
        _rating = f"{item['value'][2]}" if item["value"][2] else ""
        _rating = f"{_rating} {item['value'][3]}" if item["value"][3] else _rating
        _notes = f"{item['value'][4]}" if item["value"][4] else ""
        _duration = f"{item['value'][6]}" if item["value"][6] else ""
        # ~ typer.echo(f" - {item['value'][0]}/{item['key']} | {_status} {_rating}")
        _user = f"{user}:"
        table.add_row(
            f"{item['value'][0]}/{item['key'].replace(_user, '')}",
            _status,
            _rating,
            _notes,
            f"{', '.join(item['value'][5])}",
            _duration,
        )
    return table



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  SPDX-License-Identifier: GPL-3.0-only
#  Copyright 2022 drad <sa@adercon.com>

import logging
from os import environ, getenv, path

import pkg_resources
import tomli

from mpbroker.config.models import UserConfigBase

logger = logging.getLogger("default")

user_config_file_name = "user_config.toml"


def env_strtobool(val: str) -> bool:
    if val in ["true", "'true'"]:
        return True
    return False


def csv_to_set(in_str: str) -> set:

    return set(map(int, in_str.split(",")))


def load_config(cfg_file: str = None) -> dict:
    """
    Load config data from toml config file
    """

    conf = {}
    try:
        with open(cfg_file, "rb") as f:
            conf = tomli.load(f)
    except FileNotFoundError as e:
        print(f"File not found: {e}")
    except Exception as e:
        print(f"Other Exception: {e}")
    return conf


APP_VERSION = pkg_resources.get_distribution("mpbroker").version
APP_NAME = pkg_resources.get_distribution("mpbroker").project_name

# get user config file (e.g. ~/.config/{APP_NAME}/{user_config_file_name)
_user_config_file = None
if environ.get("MPB_CONFIG_HOME"):
    _user_config_file = path.join(environ.get("MPB_CONFIG_HOME"), user_config_file_name)
elif environ.get("XDG_CONFIG_HOME"):
    _user_config_file = path.join(
        environ.get("XDG_CONFIG_HOME"), APP_NAME, user_config_file_name
    )
elif environ.get("APPDATA"):
    _user_config_file = path.join(
        environ.get("APPDATA"), APP_NAME, user_config_file_name
    )
else:
    _user_config_file = path.join(
        environ["HOME"], ".config", APP_NAME, user_config_file_name
    )

user_cfg = UserConfigBase.parse_obj(load_config(_user_config_file))

# ENV Based Variables
DEPLOY_ENV = getenv("DEPLOY_ENV", "prd")

# logging
LOG_TO = getenv("LOG_TO", "console").split(",")
APP_LOGLEVEL = getenv("LOG_LEVEL", "WARNING")

# List of all databases (used by db_init to ensure all are created)
DATABASES = ["media", "injest_logs"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  SPDX-License-Identifier: GPL-3.0-only
#  Copyright 2022 drad <sa@adercon.com>

from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class IngestLogReason(str, Enum):
    """
    Ingest Log Reasons.
    """

    ok = "ok"
    metadata_extract_issue = "metadata extraction issue"
    updated = "updated"
    already_exists = "already exists"


class IngestLogStatus(str, Enum):
    """
    Ingest Log Statuses.
    """

    ok = "ok"
    fail = "fail"
    issue = "issue"


class IngestLogBase(BaseModel):
    """
    Ingest Log.
    """

    batchid: str = None  # ingest batch-id (YYYY-MM-DD_HHMMSS.nnnn) used to group an ingest log set.
    source: str = None  # source
    status: IngestLogStatus = None  # ok, fail, etc.
    reason: Optional[IngestLogReason] = None  # metadata_issue
    message: Optional[str] = None  # metadata extraction failed with error: xxxx


class IngestLog(IngestLogBase):
    """
    Extended (added by backend logic)
    """

    # note: this needs to be set/overwrote on result instantiation as using
    #  datetime.now() here will only get you now of when worker was started.
    created: datetime = datetime.now()
    updated: datetime = datetime.now()

    creator: Optional[str] = None
    updator: Optional[str] = None

    class Config:
        json_encoders = {datetime: lambda v: v.timestamp()}


class IngestResultType(str, Enum):
    """
    Ingest Result types.
    """

    ok = "ok, processed successfully"
    updated = "document updated"
    fail_exists = "failure, document already exists"
    extract_metadata_issue = "extract metadata issue"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  SPDX-License-Identifier: GPL-3.0-only
#  Copyright 2022 drad <sa@adercon.com>

from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel


class MediaPlayStatus(str, Enum):
    """
    Media play statuses.
    """

    new = "new"
    played = "played"
    watched = "watched"


class MediaPlayRating(str, Enum):
    """
    Media play Ratings.
    """

    unwatchable = 0
    horrible = 1
    bad = 2
    ok = 3
    good = 4
    excellent = 5


class MediaPlayHistory(BaseModel):
    """
    Media item Play History info.
    """

    base: str = None  # the base (directory) of the media_item.
    player: str = None  # the player used.
    start: datetime = None  # when play started.
    end: datetime = None  # when play ended.
    client: str = None  # the host/client:user who played the item.

    class Config:
        json_encoders = {datetime: lambda v: v.timestamp()}


class MediaPlay(BaseModel):
    """
    Media item Play info.
    """

    status: MediaPlayStatus = MediaPlayStatus.new.value  # the play status of the item.
    rating: Optional[MediaPlayRating] = None  # play rating.
    rating_notes: Optional[str] = None  # notes about rating.
    notes: Optional[str] = None  # notes about the play (e.g. watched 1, 2, 3).
    history: Optional[List[MediaPlayHistory]] = None  # play history info.


class MediaMetadata(BaseModel):
    """
    Media Metadata.
    """

    file_size: str = None  # filesize in human readable format (569 MiB, 1.1 GiB)
    file_type: str = None  # file type (video/H265)
    file_format: str = None  # file format (Matroska)
    encoding: str = None  # encoding (x265)
    duration: str = None  # duration in human readable format (1 h 52 min, 2 h 48 min)
    resolution: str = None  # resulution in width x height format (720 x 480)
    aspect_ratio: str = None  # display aspect ratio (16:9)
    audio_format: str = None  # audio format (AAC)
    audio_sampling: str = None  # audio sample rate (48000)


class MediaBase(BaseModel):
    """
    A media item.
    """

    name: str = None  # name of item (e.g. Duck_Dynasty_S1_D2.mkv)
    base: str = None  # base location of item (e.g. /opt/media)
    directory: str = None  # directory item is in (e.g. shows/Duck_Dynasty)
    sources: List[
        str
    ] = None  # list of sources the item is available at (e.g. gaz, bob, festus, etc.)
    media_type: Optional[
        str
    ] = None  # media type of media item (e.g. .mkv, .mp4, etc. - the file extension)
    notes: Optional[
        str
    ] = None  # notes about the media item (e.g. video has defect aroudn 33m into movie)
    play: Optional[MediaPlay] = None  # play info for an item.
    metadata: Optional[
        MediaMetadata
    ] = None  # metadata such as file size, duration, etc.


class MediaExt(MediaBase):
    """
    Extended (added by backend logic)
    """

    # note: this needs to be set/overwrote on result instantiation as using
    #  datetime.now() here will only get you now of when worker was started.
    created: datetime = datetime.now()
    updated: datetime = datetime.now()

    creator: Optional[str] = None
    updator: Optional[str] = None

    class Config:
        json_encoders = {datetime: lambda v: v.timestamp()}


class Media(MediaExt):
    """
    The media item.
    """

    doc_id: str = None  # doc id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  SPDX-License-Identifier: GPL-3.0-only
#  Copyright 2022 drad <sa@adercon.com>

from typing import List

from pydantic import BaseModel


class UserConfigDefaults(BaseModel):
    """
    Defaults for UserConfig.
    """

    source: str = None
    user: str = None
    base: str = None


class UserConfigIngest(BaseModel):
    """
    User Config Ingest.
    """

    file_types: List[str] = None


class UserConfigDatabase(BaseModel):
    """
    User Config database info.
    """

    db_uri: str = None


class UserConfigBase(BaseModel):
    """
    User config base.
    """

    player: str = None
    use_pager: bool = False

    database: UserConfigDatabase = None
    defaults: UserConfigDefaults = None
    ingest: UserConfigIngest = None
    source_mappings: dict = None



# -*- coding: utf-8 -*-
#  SPDX-License-Identifier: GPL-3.0-only
#  Copyright 2022 drad <sa@adercon.com>
#

import pycouchdb
import requests
import typer

from mpbroker.config.config import DATABASES, user_cfg
from mpbroker.utils import db_not_available

server = pycouchdb.Server(user_cfg.database.db_uri)

app = typer.Typer(help="[Somewhat] useful tools!")


@app.command()
def db_init():
    """
    Perform DB setup tasks: create databases, views, etc.
    NOTICE: this command is safe to use on a database that already has data.
    """
    from mpbroker.config.db_init_design_docs import views

    typer.echo("Creating Databases:")
    for db in DATABASES:
        try:
            server.create(db)
            typer.secho(f" ✓ {db} - created", fg=typer.colors.GREEN)
        except pycouchdb.exceptions.Conflict:
            typer.secho(f" ✗ {db} - skipped (already exists)", fg=typer.colors.WHITE)
        except requests.exceptions.ConnectionError:
            db_not_available()
        except Exception as e:
            typer.secho(
                f"Other error occurred during create databases: {e}",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
            raise typer.Exit()

    typer.echo("Creating Views:")
    try:
        for view in views:
            # NOTE: we take the approach of deleting all views first and then create all to ensure views are updated.
            db = server.database(view[0])
            # drop view
            db.delete(view[1])
            typer.secho(
                f" ✓ {view[0]}/{view[1]['_id']} - deleted", fg=typer.colors.GREEN
            )
            # create view
            db.save(view[1])
            typer.secho(
                f" ✓ {view[0]}/{view[1]['_id']} - created", fg=typer.colors.GREEN
            )

    except requests.exceptions.ConnectionError:
        db_not_available()
    except pycouchdb.exceptions.Conflict:
        typer.secho(
            f" ✗ {view[0]}/{view[1]['_id']} - skipped (already exists)",
            fg=typer.colors.WHITE,
        )
    except Exception as e:
        typer.secho(
            f"Other error occurred during create views: {e}",
            fg=typer.colors.RED,
            bold=True,
            err=True,
        )
        raise typer.Exit()


@app.command()
def move_media(
    from_user: str = typer.Option(
        user_cfg.defaults.user,
        "--from-user",
        help="the user account to change",
    ),
    to_user: str = typer.Option(
        user_cfg.defaults.user,
        "--to-user",
        help="the user account to change to",
    ),
    limit: int = typer.Option(
        0,
        "--limit",
        help="limit number of media items to process (0 means no limit)",
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        help="perform move without making any changes",
    ),
):
    """
    Move media (add username prefix to media in your library)
    """

    # show summary and confirm.
    typer.secho(
        "WARNING: this will move all media to the user listed below!",
        fg=typer.colors.BLACK,
        bg=typer.colors.RED,
        bold=True,
    )
    typer.secho(
        f"""\tfrom user: {from_user}
\tto user:   {to_user}
\tdb-uri:    {user_cfg.database.db_uri}
\tdry-run:   {dry_run}
\tlimit:     {limit if limit > 0 else 'none'}
"""
    )
    if not dry_run:
        typer.echo(
            "You may want to use the --dry-run parameter to test the changes before the final run"
        )
        typer.secho(
            "Only proceed if you know what you are doing!",
            fg=typer.colors.RED,
            bold=True,
        )
    typer.confirm("\nDo you want to continue?", abort=True)

    db = server.database("media")
    results = db.query(
        "filters/by-name",
        # group='true',
        # keys=[name],
        startkey=f"{from_user}",
        endkey=f"{from_user}\ufff0",
        as_list=True,
        # flat="key"
    )
    _processed = 0
    for _adoc in results:
        # skip any design-docs.
        if not _adoc["id"] == "_design/filters":
            typer.echo(f"- updating item with id: {_adoc['id']}")
            # ~ typer.echo(f" >> (adoc.id={_adoc['id']}): {_adoc}")
            doc = db.get(_adoc["id"])
            split_id = _adoc["id"].split(":")
            _name = None
            if len(split_id) < 2:
                _name = split_id[0]
            else:
                _name = split_id[1]
            # ~ typer.echo(f" >> (doc.id={doc['_id']}): {doc}")
            doc["_id"] = f"{to_user}:{_name}"
            if not dry_run:
                db.delete(_adoc["id"])
                del doc["_rev"]
                # ~ typer.echo(f" >> saving doc: {doc}")
                db.save(doc)
            typer.echo(f"  ─⏵ after update id: {doc['_id']}")
            _processed += 1
        if limit > 0 and _processed >= limit:
            break

    typer.secho(
        f"Total number processed: {_processed}", fg=typer.colors.MAGENTA, bold=True
    )



#
# NOTICE:
#  This file contains views created during 'db-init'.
#

views = []

views.append(
    [
        "media",
        {
            "_id": "_design/filters",
            "views": {
                "by-name": {
                    "map": "function(doc) { \n  if (doc.metadata) {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, doc.metadata.duration]);\n  } else {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, null]);\n  }\n}"
                },
                "by-name-new": {
                    "map": "function(doc) { \n  if (doc.metadata && doc.play.status == 'new') {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, doc.metadata.duration]);\n  } else if (doc.play.status == 'new') {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, null]);\n  }\n}"
                },
                "by-name-played": {
                    "map": "function(doc) { \n  if (doc.metadata && doc.play.status == 'played') {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, doc.metadata.duration]);\n  } else if (doc.play.status == 'played') {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, null]);\n  }\n}"
                },
                "by-name-watched": {
                    "map": "function(doc) { \n  if (doc.metadata && doc.play.status == 'watched') {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, doc.metadata.duration]);\n  } else if (doc.play.status == 'watched') {\n    emit(doc._id, [doc.directory, doc.play.status, doc.play.rating, doc.play.rating_notes, doc.play.notes, doc.sources, null]);\n  }\n}"
                },
                "stats_sources": {
                    "map": "function (doc) {\n  emit(doc.sources, 1);\n}",
                    "reduce": "_count",
                },
                "stats_status": {
                    "map": "function (doc) {\n  emit(doc.play.status, 1);\n}",
                    "reduce": "_count",
                },
                "stats_total": {
                    "map": "function (doc) {\n  emit(doc._id, 1);\n}",
                    "reduce": "_count",
                },
            },
        },
    ]
)

views.append(
    [
        "injest_logs",
        {
            "_id": "_design/filters",
            "views": {
                "status": {
                    "map": "function (doc) {\n  emit([doc.batchid, doc.status, doc.reason], 1);\n}",
                    "reduce": "_count",
                }
            },
            "language": "javascript",
        },
    ]
)



# -*- coding: utf-8 -*-
#  SPDX-License-Identifier: GPL-3.0-only
#  Copyright 2022 drad <sa@adercon.com>
#
# LOGGING: designed to run at INFO loglevel.

import json
import time
from datetime import datetime
from pathlib import Path
from typing import Optional

import click
import pycouchdb
import requests
import typer
import urllib3
from rich import box, print
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich import box

import mpbroker.tools
from mpbroker.config.config import APP_NAME, APP_VERSION, user_cfg
from mpbroker.models.ingest import IngestLog, IngestLogReason, IngestLogStatus
from mpbroker.models.media import Media, MediaPlay, MediaPlayRating, MediaPlayStatus
from mpbroker.utils import (
    db_not_available,
    determine_status_from_flags,
    extract_metadata,
    make_doc,
    play_item,
    results_by_name,
    results_to_table,
)

# disable InsecureRequestWarnings which come up if you are proxying couchdb through haproxy with ssl termination.
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

server = pycouchdb.Server(user_cfg.database.db_uri)
app = typer.Typer()
app.add_typer(mpbroker.tools.app, name="tools")


@app.command()
def info():
    """
    All about your Library
    """

    # NOTICE: this command is not 'user' aware, info is for complete Library.

    try:
        db = server.database("media")

        _total = db.query(
            "filters/stats_total",
            # ~ group='true',
            # ~ keys=[name],
            # ~ startkey=name,
            # ~ endkey=f"{name}\ufff0",
            as_list=True,
            # ~ flat="key"
        )
        if not _total or len(_total) < 1:
            typer.secho(
                "Your Library appears to be empty try, ingesting something!",
                fg=typer.colors.MAGENTA,
                bold=True,
            )
            raise typer.Exit()

        _status = db.query(
            "filters/stats_status",
            group="true",
            as_list=True,
        )

        _sources = db.query(
            "filters/stats_sources",
            group="true",
            as_list=True,
        )

        _sources_list = [
            f"\n  • {source['key'][0]} ({source['value']})" for source in _sources
        ]
        _new = [item for item in _status if item["key"] == MediaPlayStatus.new.value]
        _played = [item for item in _status if item["key"] == MediaPlayStatus.played.value]
        _watched = [item for item in _status if item["key"] == MediaPlayStatus.watched.value]

        _about = f"""{APP_NAME} v{APP_VERSION}
Copyright 2022 drad <sa@adercon.com>"""
        _account = f"- Database: [bold]{user_cfg.database.db_uri}[/bold]"
        _defaults = f"""- User:   {user_cfg.defaults.user}
- Source: {user_cfg.defaults.source}
- Base:   {user_cfg.defaults.base}"""
        _preferences = f"""- Use Pager: {user_cfg.use_pager}
- Player:    {user_cfg.player}"""
        _library = f"""- Total:   [bold]{_total[0]['value']}[/bold]
- New:     {_new[0]['value'] if _new else 0}
- Played:  {_played[0]['value'] if _played else 0}
- Watched: {_watched[0]['value'] if _watched else 0}
- Sources {''.join(_sources_list)}"""

        _width = 100
        print(Panel(_about, expand=True, width=_width, box=box.MINIMAL, style=""))
        print(
            Panel(
                _account, title="Account", expand=True, width=_width, style="on gray30"
            )
        )
        print(
            Panel(
                _defaults,
                title="Defaults",
                expand=True,
                width=_width,
                style="on gray30",
            )
        )
        print(
            Panel(
                _preferences,
                title="Preferences",
                expand=True,
                width=_width,
                style="on gray30",
            )
        )
        print(
            Panel(
                _library, title="Library", expand=True, width=_width, style="on gray30"
            )
        )

    except requests.exceptions.ConnectionError:
        db_not_available()
    except pycouchdb.exceptions.NotFound:
        typer.secho(
            "Your Library appears to be empty try, ingesting something!",
            fg=typer.colors.RED,
            bold=True,
        )
        raise typer.Exit()


@app.command()
def ingest(
    base: str = typer.Option(
        user_cfg.defaults.base,
        "--base",
        help="the base path to search for media items to ingest",
    ),
    source: str = typer.Option(
        user_cfg.defaults.source,
        "--source",
        help="source to use for ingest (ingested data will have this source)",
    ),
    user: str = typer.Option(
        user_cfg.defaults.user,
        "--user",
        help="user to use for ingest (ingested data will belong to this user)",
    ),
    extract_metadata_flag: bool = typer.Option(
        False,
        "--extract-metadata",
        help="extract media metadata (e.g. duration, size, encoding type, etc.) - note: will only extract on for 'new' records (use --extract-metadata-all to extract for all items)",
    ),
    extract_metadata_all_flag: bool = typer.Option(
        False,
        "--extract-metadata-all",
        help="extract metadata for all items (not just new items)"
    ),
    no_confirm: bool = typer.Option(
        False,
        "--no-confirm",
        help="do not confirm - run ingest immediately (useful if you are running via cron or a script)",
    ),
):
    """
    ingest media
    """

    console = Console()

    typer.echo(f"Scanning base ({base}) for media...")
    _user = user if len(user) > 0 else None

    _base = Path(base)
    if not _base.exists() or not _base.is_dir():
        typer.echo(f"Directory [{_base}] not found or not a directory, cannot proceed!")
        raise typer.Exit()

    all_files = []
    for ext in user_cfg.ingest.file_types:
        all_files.extend(_base.rglob(ext))

    sum_table = Table(box=box.ROUNDED, show_header=False, show_footer=False, show_lines=True, title_style="bold magenta", style="on gray30")
    sum_table.title = "Ingest Summary"
    sum_table.add_row("[bold cyan]File Types", ",".join(user_cfg.ingest.file_types), style="on gray30")
    sum_table.add_row("[bold cyan]Base", _base.as_posix(), style="on gray30")
    sum_table.add_row("[bold cyan]User", _user, style="on gray30")
    sum_table.add_row("[bold cyan]Source", source, style="on gray30")
    sum_table.add_row("[bold cyan]Extract Metadata", f"{extract_metadata_flag}", style="on gray30")
    sum_table.add_row("[bold cyan]Extract Metadata All", f"{extract_metadata_all_flag}", style="on gray30")
    sum_table.add_row("[bold cyan]Number of Items", f"{len(all_files)}", style="on gray30")

    if no_confirm is False:
        console.print(sum_table)
        typer.confirm("Do you want to continue?", abort=True)
        typer.echo("")

    _start = time.time()
    media_db = server.database("media")
    _batchid = datetime.now().strftime("%Y-%m-%d_%H:%M:S.%f")

    with typer.progressbar(all_files, label="Ingesting") as progress:
        for f in progress:
            ingest_file(
                source=source,
                filepath=f,
                base=_base.as_posix(),
                extract_metadata_flag=extract_metadata_flag,
                batchid=_batchid,
                user=_user,
                db=media_db,
            )

    _stop = time.time()
    typer.echo("")
    _width = 100
    run_table = Table(box=box.ROUNDED, show_header=False, show_footer=False, show_lines=True, style="on gray30", expand=True)
    run_table.add_row("[bold cyan]Batch Id", _batchid, style="on gray30")
    run_table.add_row("[bold cyan]Ingest Time", f"{_stop - _start}s", style="on gray30")
    ing_table = Table(box=box.ROUNDED, show_header=False, show_footer=False, show_lines=True, style="on gray30", expand=True)
    ing_table.add_column(ratio=1)
    ing_table.add_column(ratio=2)

    # get ingest_logs info
    try:
        injest_logs_db = server.database("injest_logs")
        _il_status = injest_logs_db.query(
            "filters/status",
            group="true",
            # ~ keys=['2022-05-03_12:24:35.4292'],
            startkey=[_batchid],
            endkey=[f"{_batchid}\ufff0"],
            as_list=True,
            # ~ flat="key"
        )
        for row in _il_status:
            _label = f"{row['key'][1]} ({row['key'][2]})"
            _value = f"{row['value']}"
            ing_table.add_row(f"[bold cyan]{_label}", _value, style="on gray30")

    except requests.exceptions.ConnectionError:
        db_not_available()


    print(
        Panel(
            run_table, title="Run Info", expand=True, width=_width, style="on gray30"
        )
    )
    print(
        Panel(
            ing_table, title="Ingest Info", expand=True, width=_width, style="on gray30"
        )
    )


def ingest_file(
    source: str,
    filepath: str,
    base: str,
    extract_metadata_flag: bool = False,
    batchid: str = None,
    user: str = None,
    db = None
):
    """
    Ingest a file.
    """

    # ensure base ends with /
    _base = base if base.endswith("/") else f"{base}/"
    _user = f"{user}:" if user else ""
    # directory is filepath.parent - base
    directory = str(filepath.parent).replace(_base, "")

    # create IngestLog instance.
    il = IngestLog(
        batchid=batchid,
        source=source,
        status=IngestLogStatus.ok,
        reason=IngestLogReason.ok,
        message=None,
        created=datetime.now(),
        creator=_user,
    )

    m = Media(
        doc_id=f"{_user}{filepath.name}",
        # sid=make_sid(filepath.name),
        name=filepath.name,
        base=_base,
        directory=directory,
        sources=[source],
        media_type=filepath.suffix,
        # ~ notes="",
        play=MediaPlay(),
        metadata=None,  # added later
        creator=None,
        updator=None,
    )
    _doc_id = f"{_user}{filepath.name}"

    try:
        #db = server.database("media")
        # NOTICE: this is the 'update' path, most calls should drop to the pycouchdb.exceptions.NotFound path.
        # check if doc already exists.
        _doc = db.get(_doc_id)
        # update the doc as needed.
        if source not in _doc["sources"]:
            _doc["sources"].append(source)
            il.reason = IngestLogReason.updated

        if extract_metadata_flag:
            _metadata, _error = extract_metadata(filepath)
            _doc["metadata"] = json.loads(_metadata.json()) if _metadata else None
            il.reason = IngestLogReason.updated
            if _error:
                il.status = IngestLogStatus.issue
                il.reason = IngestLogReason.metadata_extract_issue

        if (
            il.reason == IngestLogReason.updated
            or il.reason == IngestLogReason.metadata_extract_issue
        ):
            _doc["updated"] = datetime.timestamp(datetime.now())
            db.save(_doc)
        else:
            il.status = IngestLogStatus.ok
            il.reason = IngestLogReason.already_exists

    except requests.exceptions.ConnectionError:
        db_not_available()
    except pycouchdb.exceptions.NotFound:
        # new item: add to Library
        il.status = IngestLogStatus.ok
        il.reason = IngestLogReason.ok
        metadata = None
        if extract_metadata_flag:
            m.metadata, error = extract_metadata(filepath)
            # if no metadata returned check error.
            if not metadata:
                il.status = IngestLogStatus.issue
                il.reason = IngestLogReason.metadata_extract_issue

        dts = make_doc(doc=m, rename_doc_id=True)

        db.save(dts)

    except pycouchdb.exceptions.Conflict:
        il.status = IngestLogStatus.fail
        il.reason = IngestLogReason.already_exists
        il.message = f"Duplicate item not ingested: {m.directory}/{m.name}, source(s): {_doc['sources']}"

    # write the IngestLog record.
    try:
        db = server.database("injest_logs")
        db.save(
            # stopped here, need to get an _id on the il doc somewhere and go...
            make_doc(doc=il, rename_doc_id=False)
        )
    except requests.exceptions.ConnectionError:
        db_not_available()


@app.command()
def list(
    name: str,
    directory: str = typer.Option(
        None, "--directory", help="directory/partial directory to filter on"
    ),
    new: bool = typer.Option(False, "--new", "-n", help="only select new items"),
    played: bool = typer.Option(
        False, "--played", "-p", help="only select played items"
    ),
    watched: bool = typer.Option(
        False, "--watched", "-w", help="only select watched items"
    ),
    auto_play: bool = typer.Option(
        False, "--auto-play", "-a", help="auto-play the first item returned in list"
    ),
    user: str = typer.Option(
        user_cfg.defaults.user,
        "--user",
        help="only show for user (leave empty to use default user)",
    ),
):
    """
    List media
    """

    _status = determine_status_from_flags(new=new, played=played, watched=watched)
    _results = results_by_name(
        name=name, user=user, status=_status, filter_directory=directory
    )

    if auto_play:
        try:
            db = server.database("media")
            # play first item
            _doc = db.get(_results[0]["id"])
            _fmt_item = typer.style(
                f"{_doc['directory']}/{_doc['name']}", bg=typer.colors.RED, bold=True
            )
            typer.echo(f"Auto-Play of item: {_fmt_item}")
            play_item(doc=_doc, name=name, user=user)
        except requests.exceptions.ConnectionError:
            db_not_available()
    else:
        table = results_to_table(_results, name=name, user=user, directory=directory)
        console = Console()
        if user_cfg.use_pager:
            with console.pager(styles=True):
                console.print(table)
        else:
            console.print(table)


@app.command()
def play(
    name: str,
    user: str = typer.Option(
        user_cfg.defaults.user,
        "--user",
        help="user to use for playing",
    ),
):
    """
    Play media
    """

    _base = Path(name)
    media_item = _base.name
    # typer.echo(f"Playing item [{name}/{media_item}]")

    # lookup item to get source.
    try:
        db = server.database("media")
        _id = f"{user}:{media_item}"
        # ~ typer.echo(f"- getting media item with id={_id} of name={name}")
        # add 'play next' logic; I think we should create a 'get_or_select_media_item' function here which tries to get a doc, if more then one is found it prints list and lets user choose, then returns doc
        _doc = db.get(_id)

        play_item(doc=_doc, name=name, user=user)

    except requests.exceptions.ConnectionError:
        db_not_available()
    # ~ except pycouchdb.exceptions.NotFound:
    # ~ typer.secho(
    # ~ f"Media item {name} not found for user {user}.",
    # ~ fg=typer.colors.RED,
    # ~ bold=True,
    # ~ )
    # ~ raise typer.Exit()


@app.command()
def remove(
    name: str,
    directory: str = typer.Option(
        None, "--directory", help="directory/partial directory to filter on"
    ),
    new: bool = typer.Option(False, "--new", "-n", help="only select new items"),
    played: bool = typer.Option(
        False, "--played", "-p", help="only select played items"
    ),
    watched: bool = typer.Option(
        False, "--watched", "-w", help="only select watched items"
    ),
    user: str = typer.Option(
        user_cfg.defaults.user,
        "--user",
        help="only show for user (leave empty to use default user)",
    ),
):
    """
    Remove media
    """

    _status = determine_status_from_flags(new=new, played=played, watched=watched)

    try:
        results = results_by_name(
            name=name, user=user, status=_status, filter_directory=directory
        )
        table = results_to_table(results, name=name, user=user)
        console = Console()
        # NOTE: no pagination for delete display
        console.print(table)
        _action = typer.style("Remove ALL", bg=typer.colors.RED, bold=True)
        typer.confirm(
            f"Are you sure you want to {_action} items listed above?", abort=True
        )
        _removed = 0
        db = server.database("media")
        with typer.progressbar(results) as progress:
            for item in progress:
                db.delete(item["id"])
                _removed += 1

        typer.secho(
            f"Deleted {_removed} items from your Library!",
            fg=typer.colors.MAGENTA,
            bold=True,
        )

    except requests.exceptions.ConnectionError:
        db_not_available()
    except pycouchdb.exceptions.NotFound:
        typer.secho(
            f"Media item {name} not found for user {user}.",
            fg=typer.colors.RED,
            bold=True,
        )
        raise typer.Exit()


@app.command()
def update(
    name: str,
    directory: str = typer.Option(
        None, "--directory", help="directory/partial directory to filter on"
    ),
    new: bool = typer.Option(False, "--new", "-n", help="only select new items"),
    played: bool = typer.Option(
        False, "--played", "-p", help="only select played items"
    ),
    watched: bool = typer.Option(
        False, "--watched", "-w", help="only select watched items"
    ),
    user: str = typer.Option(
        user_cfg.defaults.user,
        "--user",
        help="only show for user (leave empty to use default user)",
    ),
):
    """
    Update media
    """

    _status = determine_status_from_flags(new=new, played=played, watched=watched)

    # @TODO: flake8 complexity of this method is 20, need to refactor to bring it down to 18 at most.
    try:
        results = results_by_name(
            name=name, user=user, status=_status, filter_directory=directory
        )
        table = results_to_table(results, name=name, user=user)
        console = Console()
        # NOTE: no pagination for update display.
        console.print(table)

        typer.confirm("Do you want to continue?", abort=True)
        _rating_list = [str(i) for i in MediaPlayRating._value2member_map_]
        _rating_list.append("")
        _status = None

        if typer.confirm("  Update Status?"):
            _status = typer.prompt(
                "    New Status",
                default=MediaPlayStatus.watched.value,
                type=click.Choice([str(i) for i in MediaPlayStatus._value2member_map_]),
            )
        _update_rating = typer.confirm("  Update Rating?")
        if _update_rating:
            _rating = typer.prompt(
                "    New Rating (blank to clear)",
                default="",
                type=click.Choice(_rating_list),
            )
        _update_rating_notes = typer.confirm("  Update Rating Notes?")
        if _update_rating_notes:
            _rating_notes = typer.prompt(
                "  New Rating Notes (blank to clear)",
                default="",
                show_default=False,
            )
        _update_notes = typer.confirm("  Update Notes?")
        if _update_notes:
            _notes = typer.prompt(
                "    New Notes (blank to clear)",
                default="",
                show_default=False,
            )
        _extract_metadata = typer.confirm("  Extract Metadata for item?")
        # iterate over docs and update accordingly.
        _updated = 0
        db = server.database("media")
        _errors = []
        with typer.progressbar(results) as progress:
            for item in progress:
                _doc = db.get(item["id"])
                _dirty = False
                if _status:
                    _doc["play"]["status"] = _status
                    _dirty = True
                if _update_rating:
                    if _rating:
                        _doc["play"]["rating"] = int(_rating)
                    else:
                        _doc["play"]["rating"] = None
                    _dirty = True
                if _update_rating_notes:
                    if _rating_notes:
                        _doc["play"]["rating_notes"] = _rating_notes
                    else:
                        _doc["play"]["rating_notes"] = None
                    _dirty = True
                if _update_notes:
                    if _notes:
                        _doc["play"]["notes"] = _notes
                    else:
                        _doc["play"]["notes"] = None
                    _dirty = True
                if _extract_metadata:
                    _filepath = f"{_doc['base']}{_doc['directory']}/{_doc['name']}"
                    # ~ typer.echo(f"    ::> {_filepath}")
                    metadata, error = extract_metadata(_filepath)
                    if metadata and not error:
                        _doc["metadata"] = json.loads(metadata.json())
                        _dirty = True
                    else:
                        _errors.append(f"  ✦ {error}")
                if _dirty:
                    _updated += 1
                    _doc["updated"] = datetime.timestamp(datetime.now())
                    _doc["updator"] = user
                    db.save(_doc)
        if len(_errors) > 0:
            typer.secho("\nMetadata Extraction Issues:", fg=typer.colors.MAGENTA)
            typer.secho("\n".join(_errors), fg=typer.colors.YELLOW)
        typer.secho(
            f"Updated {_updated} items from your Library!",
            fg=typer.colors.MAGENTA,
            bold=True,
        )

    except requests.exceptions.ConnectionError:
        db_not_available()
    except pycouchdb.exceptions.NotFound:
        typer.secho(
            f"Media item {name} not found for user {user}.",
            fg=typer.colors.RED,
            bold=True,
        )
        raise typer.Exit()


def version_callback(value: bool):
    if value:
        typer.echo(f"{APP_NAME} {APP_VERSION}")

        raise typer.Exit()


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        help="Show application version and exit",
    ),
):

    pass


if __name__ == "__main__":
    app()
